"""Functions for cloudwatch"""

from dai_python_commons.dai_error import DaiInputError


class CWUtils:  # pylint: disable=too-few-public-methods
    """
    Class that provides helpful CloudWatch utilities.
    """

    @staticmethod
    def put_lambda_metric(cw_client, metric_name: str, dai_code: str, function_name: str):
        """
        Increase counter for our dai-code metric for its kind of action in cloudwatch.

        :param cw_client: A boto3.client("cloudwatch") to use to write the logs
        :param metric_name: Which counter to increase
        :param dai_code: The dai_code (e.g. `tr-tms-obs-0008`) which will be used for setting the dimension of the metric
        :param function_name: The lambda name which will be used for setting the namespace of the metric
        """
        allowed_metric_names = ["invocation", "error", "validationError"]
        if metric_name not in allowed_metric_names:
            raise DaiInputError(f"{metric_name} not in {allowed_metric_names}")

        cw_client.put_metric_data(
            Namespace=f"DAI/Lambda/{function_name}",
            MetricData=[
                {
                    "MetricName": metric_name,
                    "Value": 1,
                    "Unit": "Count",
                    "Dimensions": [{"Name": "daiCode", "Value": dai_code}],
                }
            ],
        )



"""Useful functions for date/time"""
import datetime
from typing import List


DATE_FORMAT = "%Y-%m-%d"


class TimeUtils:
    """
    Class that provides useful functions for dates and times
    """

    @staticmethod
    def get_dates_in_time_interval(start_date: datetime.date, end_date: datetime.date) -> List[datetime.date]:
        """
        Get all dates between start and end date. It includes the start_date, but not the end_date.

        :param start_date: The start date
        :param end_date: The end date
        :return:
        """
        delta = end_date - start_date  # as timedelta
        days_between_start_end = [start_date + datetime.timedelta(days=i) for i in range(delta.days)]

        return days_between_start_end

    @staticmethod
    def generate_partition_folder_names(dates: List[datetime.date], partition_column="dt") -> List[str]:
        """
        Generates the partition folder names for a list of dates, eg for [2020-01-01, 2020-01-02] and partition column
        dt it generates ["dt=2020-01-01", "dt=2020-01-02"]

        :param dates: List of dates
        :param partition_column: Name of the partition column
        :return: List of folder names for the given dates
        """
        formatted_dates = [d.strftime(DATE_FORMAT) for d in dates]

        return [f"{partition_column}={d}" for d in formatted_dates]

    @staticmethod
    def string2date(the_date: str, date_format: str = DATE_FORMAT) -> datetime.date:
        """
        Convert a string to date

        :param the_date: date provided as string
        :param date_format: the format in which the date is provided
        :return: The date as a datetime.date
        """

        return datetime.datetime.strptime(the_date, date_format).date()

    @staticmethod
    def date2string(the_date: datetime.date, date_format: str = DATE_FORMAT) -> str:
        """
        Convert a string to date

        :param the_date: date provided as datetime.date
        :param date_format: the desired format
        :return: The date as a string
        """

        return the_date.strftime(date_format)

    @staticmethod
    def get_yesterday() -> datetime.date:
        """
        Gets yesterday's date as datetime.date

        :return: Yesterday's date
        """
        yesterday_datetime = datetime.date.today() - datetime.timedelta(1)

        return yesterday_datetime



"""Functionality to manipulate S3 events from EventBridge"""
import os
from typing import Any, Dict

from dai_python_commons.dai_error import MissingInputError, IncorrectInputValueError


class EventBridgeS3Event:
    """
    Class that provides useful functions to handle Event Bridge S3 events
    """

    @staticmethod
    def validate_config_dict(event_dict: Dict[str, Any]) -> None:
        """
        Validates that essential fields are present in an event bridge s3 event

        :param event_dict: The S3 EventBridge event given as a dictionary
        :return: Nothing, but it raises an exception if the event does not pass the validation
        """
        missing_fields = [field for field in ["source", "detail"] if field not in event_dict]
        if len(missing_fields) > 0:
            raise MissingInputError(missing_fields, "Incorrect event bridge s3 event, fields are missing.")

        # Make sure this is an s3 event
        if event_dict.get("source") != "aws.s3":
            raise IncorrectInputValueError({"source": event_dict["source"]}, "The source should be aws.s3")

        # check that the important fields are there
        missing_fields = [field for field in ["bucket", "object"] if field not in event_dict["detail"]]
        if len(missing_fields) > 0:
            raise MissingInputError(missing_fields, "Missing needed information in s3 event, detail field.")

        # check that the bucket name is present
        if "name" not in event_dict["detail"]["bucket"]:
            raise MissingInputError(["name"], "Missing 'name' field in detail.bucket")

        # check that the key is present
        if "key" not in event_dict["detail"]["object"]:
            raise MissingInputError(["key"], "Missing 'key' field in detail.object")

    def __init__(self, event_dict: Dict[str, Any]):
        """
        :param event_dict: The S3 EventBridge event given as a dictionary
        """
        EventBridgeS3Event.validate_config_dict(event_dict)

        self.s3_bucket = event_dict["detail"]["bucket"]["name"]
        self.s3_key = event_dict["detail"]["object"]["key"]
        self.s3_prefix = os.path.dirname(self.s3_key)
        self.original_event = event_dict

    def get_s3_path(self) -> str:
        """
        :return: Full s3 path for the object that generated the event
        """
        return f"s3://{self.s3_bucket}/{self.s3_key}"



"""Useful functions for s3 buckets"""
from __future__ import annotations

import datetime
import re
from typing import Iterable, List

import loguru
from botocore.response import StreamingBody
from mypy_boto3_s3.client import S3Client
from mypy_boto3_s3.type_defs import DeleteTypeDef, ObjectIdentifierTypeDef

from dai_python_commons import dai_error

# pylint: disable=no-member

class S3Utils:
    """
    Class provides functions to modify s3 buckets
    """

    RAW_BUCKET_PATTERN = re.compile(r"se-sl-tf-dai-[^-]*-data-raw")

    @staticmethod
    def delete_objects_by_prefix(
        boto_s3_client: S3Client,
        bucket_name: str,
        prefix: str,
        logger: loguru.Logger,
    ) -> int:
        """
        Deletes all objects with given prefix from a bucket

        :param boto_s3_client: S3 client to be used
        :param bucket_name: Name of the bucket
        :param prefix: Prefix of objects that should be removed
        :param logger: The logger
        :return: Number of objects deleted
        """
        logger.info(f"Deleting objects with prefix {prefix} from bucket {bucket_name}")

        paginator = boto_s3_client.get_paginator("list_objects_v2")
        page_iterator = paginator.paginate(Bucket=bucket_name, Prefix=prefix)

        # gather all objects to be deleted
        to_delete = []
        for item in page_iterator.search("Contents"):
            if not item:
                break
            to_delete.append(dict(Key=item["Key"]))

        # delete the objects
        return S3Utils.delete_objects(
            boto_s3_client=boto_s3_client,
            bucket_name=bucket_name,
            to_delete=to_delete,
            logger=logger,
        )

    @staticmethod
    def delete_objects(
        boto_s3_client: S3Client,
        bucket_name: str,
        to_delete: List[ObjectIdentifierTypeDef],
        logger: loguru.Logger,
    ) -> int:
        """
        Deletes the specified objects in `to_delete` and returns the number of deleted objects.

        `to_delete` should be in the following format:

            [
                {
                    'Key': 'prefix'
                }
            ]

        :param boto_s3_client: S3 client to be used
        :param bucket_name: Name of the bucket
        :param to_delete: list of objects to delete
        :param logger: The logger
        :return: Number of objects deleted
        """

        start_index = 0
        while start_index < len(to_delete):
            end_index = min(len(to_delete), start_index + 1000)
            batch_to_delete = DeleteTypeDef( Objects=to_delete[start_index:end_index] )
            logger.debug(f"Deleting {batch_to_delete}")
            response = boto_s3_client.delete_objects(
                Bucket=bucket_name, Delete=batch_to_delete
            )

            # If something went wrong raise an error
            if "Errors" in response and len(response["Errors"]) > 0:
                raise dai_error.DaiS3DeleteObjectError(
                    s3_bucket=bucket_name, error_infos=response["Errors"]
                )

            start_index = end_index

        logger.info(f"Number of objects deleted: {len(to_delete)}")

        return len(to_delete)

    @staticmethod
    def iter_file_paths_in_prefix(
        boto_s3_client: S3Client,
        bucket_name: str,
        prefix: str,
        logger: loguru.Logger,
        max_size: int,
    ) -> Iterable[List[ObjectIdentifierTypeDef]]:
        """
        Returns a Iterable over lists of dicts representing file paths in the format:

            [
                {
                    'Key': 'prefix'
                }
            ]

        NOTE: Files larger than max_size are never returned!

        :param boto_s3_client: S3 client to be used
        :param bucket_name: Name of the bucket
        :param prefix: Prefix where to list files
        :param logger: The logger
        :param max_size: The size in bytes that limits how many files are returned in one iteration
        :yields: lists of file path dicts
        """

        paginator = boto_s3_client.get_paginator("list_objects_v2")
        prefix = prefix if prefix.endswith("/") else f"{prefix}/"
        available_file_paths = []
        for item in paginator.paginate(
            Bucket=bucket_name, Prefix=prefix, Delimiter="/"
        ).search("Contents"):
            # Ignore directories, i.e. suffix '/'
            if item and not item["Key"].endswith("/"):
                available_file_paths.append({"Key": item["Key"], "Size": item["Size"]})

        prefixes = []
        size = 0
        for item in available_file_paths:
            if item["Size"] + size > max_size and len(prefixes) > 0:
                logger.debug(
                    f'In "{bucket_name}/{prefix}" yields partial results: {prefixes} at size {size}'
                )
                yield prefixes
                size = 0
                prefixes.clear()

            if item["Size"] < max_size:
                prefixes.append({"Key": item["Key"]})
                size += item["Size"]

        logger.debug(f'In "{bucket_name}/{prefix}" yields last results: {prefixes}')

        yield prefixes

    @staticmethod
    def file_paths_in_prefix(
        boto_s3_client: S3Client,
        bucket_name: str,
        prefix: str,
        logger: loguru.Logger,
    ) -> List[ObjectIdentifierTypeDef]:
        """
        Returns a list of dicts representing file paths in the format:

            [
                {
                    'Key': 'prefix'
                }
            ]

        :param boto_s3_client: S3 client to be used
        :param bucket_name: Name of the bucket
        :param prefix: Prefix where to list files
        :param logger: The logger
        :return: A list of file path dicts
        """
        paginator = boto_s3_client.get_paginator("list_objects_v2")
        prefix = prefix if prefix.endswith("/") else f"{prefix}/"
        file_paths = []
        for item in paginator.paginate(
            Bucket=bucket_name, Prefix=prefix, Delimiter="/"
        ).search("Contents"):
            # Ignore directories, i.e. suffix '/'
            if item and not item["Key"].endswith("/"):
                file_paths.append({"Key": item["Key"]})

        logger.debug(f'In "{bucket_name}/{prefix}" found: {file_paths}')

        return file_paths

    @staticmethod
    def is_raw_bucket(bucket_name: str) -> bool:
        """Checks that bucket_name matches RAW_BUCKET_PATTERN regex"""
        return S3Utils.RAW_BUCKET_PATTERN.match(bucket_name) is not None

    @staticmethod
    def get_last_modified(
        boto_s3_client: S3Client,
        bucket_name: str,
        key: str,
        logger: loguru.Logger,
    ) -> datetime.datetime:
        """
        Gets the last modified stamp for a object stored in s3

        :param boto_s3_client: Boto3 s3 client
        :param bucket_name: Name of s3 bucket
        :param key: Object key
        :param logger: Logger object
        :return: Last modified value
        """
        response = boto_s3_client.head_object(Bucket=bucket_name, Key=key)
        if response["ResponseMetadata"]["HTTPStatusCode"] != 200:
            error_info = response.get("Error", {})
            logger.error(
                f"Unable to obtain LastModified for s3://{bucket_name}/{key}. Error info: {error_info}"
            )
            raise Exception(error_info.get("Message", ""))

        return response["LastModified"]

    @staticmethod
    def read_s3_file(s3_client: S3Client, bucket_name: str, key: str) -> StreamingBody:
        """
        Get content of an S3 object

        :param s3_client: Boto3 s3 client
        :param bucket_name: Name of s3 bucket
        :param key: Object key
        :return: Http response body
        """
        response = s3_client.get_object(Bucket=bucket_name, Key=key)
        content = response["Body"]

        return content



"""Functionality to modify parquet files in an S3 bucket"""
from __future__ import annotations

import gc
from dataclasses import dataclass
from typing import Optional

import boto3
import loguru
import pyarrow.parquet as pq
import s3fs

from dai_python_commons.s3_utils import S3Utils


@dataclass
class ParquetFileLocation:
    """
    Class that contains information about parquet files location in S3.

    args:
        source_bucket (str): The bucket where the parquet files to be merged are located.
        source_prefix (str): The prefix (ie 'folder') where the parquet files are located.
        destination_bucket (str): The bucket where the larger parquet file should be located. It can be the same as the
        source bucket.

        destination_prefix (str): The prefix (ie 'folder') where the merged file should be written
        compression (str): Type of compression. Accepted values are 'NONE', 'SNAPPY', 'GZIP', 'BROTLI', 'LZ4', 'ZSTD'
        remove_files_at_destination (bool): whether the files at the destination folder should be removed before
        writing the merged parquet file

        keep_source_files (bool): whether the source files should be kept after merge
        maximum_file_size (int): only parse sets of files which totals (in Bytes) under this limit
    """
    # pylint: disable=R0902
    source_bucket: str
    source_prefix: str
    destination_bucket: str = ""
    destination_prefix: str = ""
    compression: Optional[str] = "SNAPPY"
    remove_files_at_destination: bool = False
    keep_source_files: bool = True
    total_files_size: int = 1024 * 1024 * 1024   # 1GB


class ParquetUtils:
    """
    Class that provides functionality for manipulating parquet files
    """
    # pylint: disable=no-member
    VALID_COMPRESSION_TYPES = {'NONE', 'SNAPPY', 'GZIP', 'BROTLI', 'LZ4', 'ZSTD'}

    @staticmethod
    def s3_merge_files_in_place(boto3_session: boto3.Session,
                                parquet_file_location: ParquetFileLocation,
                                logger: loguru.Logger,
                                compression: Optional[str] = "SNAPPY",
                                keep_source_files: bool = False) -> int:
        """
         Merge many small parquet files into one larger parquet file. In place.

        :param boto3_session: Boto3 session
        :param parquet_file_location: s3 bucket and prefix location where parquet files will be merged
        :param logger: The logger
        :param compression: Type of compression. Accepted values are 'NONE', 'SNAPPY', 'GZIP', 'BROTLI', 'LZ4', 'ZSTD'
        :param keep_source_files: True to keep the source files, False to delete them.
        :return:
        """
        data_path = f"s3://{parquet_file_location.source_bucket}/{parquet_file_location.source_prefix}"
        logger.info(f"Merging files from {data_path} to {data_path}, compression: {compression}")

        # this is a bit of hack, one should probably find a better way to do this
        botoc_session = boto3_session._session  # pylint: disable=protected-access
        f_s = s3fs.S3FileSystem(session=botoc_session)
        botoc_s3_client = botoc_session.create_client("s3")
        num_rows = 0

        file_paths_iter = S3Utils.iter_file_paths_in_prefix(boto_s3_client=botoc_s3_client,
                                                  bucket_name=parquet_file_location.source_bucket,
                                                  prefix=parquet_file_location.source_prefix,
                                                  logger=logger,
                                                  max_size=parquet_file_location.total_files_size)
        for file_paths in file_paths_iter:
            if len(file_paths) <2:
                logger.info(f"file paths is less than 2 files, no reason to merge: {file_paths}")
                continue
            logger.info(f"Reading and merging {file_paths}")

            full_s3_paths = [f's3://{parquet_file_location.source_bucket}/{file_path["Key"]}' for file_path in file_paths]

            # partitioning=None ensures we do not add a column dt when parsing
            # our parquet files that usually have a "hive-like" partitioned dt
            # path, which we do not wish to add to the set when consolidating.
            # The type in the interface is wrong, source code does handle the None
            pq_table = pq.read_table(full_s3_paths, filesystem=f_s, partitioning=None)  # type: ignore
            num_rows += pq_table.num_rows
            logger.debug(f"Shape of the table {pq_table.shape}")

            logger.debug(f"Writing to the destination {data_path}")
            try:
                pq.write_to_dataset(pq_table, data_path, filesystem=f_s, compression=compression, flavor="spark")
                logger.info(f"Done merging, {pq_table.num_rows} rows were written at {data_path}")
                if not keep_source_files:
                    logger.info('Removing source files')
                    logger.debug(f'Removing these files {file_paths}')
                    S3Utils.delete_objects(
                        boto_s3_client=botoc_s3_client,
                        bucket_name=parquet_file_location.source_bucket,
                        to_delete=file_paths,
                        logger=logger
                    )
            except Exception:
                logger.exception(f'Caught error when trying to merge parquet files: {file_paths}')
                raise

            # Explicitly clear data from memory
            del pq_table
            gc.collect()

        if num_rows == 0:
            logger.warning(f"No files found at {data_path}, nothing to merge")

        return num_rows

    @staticmethod
    def s3_merge_files(boto3_session: boto3.Session, parquet_file_location: ParquetFileLocation,
                       logger: loguru.Logger) -> int:
        """
        Merge many small parquet files into one larger parquet file. From source to destination.
        Exception will be raised if source is equals to destination.

        :param boto3_session: Boto3 session
        :param parquet_file_location: ParquetFileLocation contains info about the files location in the s3 bucket
        :param logger: The logger
        :return: Number of rows in the parquet file
        """
        source_bucket = parquet_file_location.source_bucket
        source_prefix = parquet_file_location.source_prefix
        remove_files_at_destination = parquet_file_location.remove_files_at_destination

        ParquetUtils._source_and_destination_not_same(parquet_file_location)

        source_data_path = f"s3://{parquet_file_location.source_bucket}/{parquet_file_location.source_prefix}"
        destination_data_path = f"s3://{parquet_file_location.destination_bucket}/{parquet_file_location.destination_prefix}"

        logger.info(f"Merging files from {source_data_path} to {destination_data_path},"
                    f" compression: {parquet_file_location.compression}, "
                    f"remove_files_at_destination={parquet_file_location.remove_files_at_destination}")

        # check if there are any files present
        s3_client = boto3_session.client('s3')
        file_paths = S3Utils.file_paths_in_prefix(boto_s3_client=s3_client,
                                                  bucket_name=source_bucket,
                                                  prefix=source_prefix,
                                                  logger=logger)
        if len(file_paths) == 0:
            logger.warning(f"No files found at {source_data_path}, nothing to merge")
            return 0

        if remove_files_at_destination:
            ParquetUtils._remove_files(parquet_file_location, destination_data_path, logger, s3_client)

        logger.debug(f"Reading data from {source_data_path}")
        # this is a bit of hack, one should probably find a better way to do this
        f_s = s3fs.S3FileSystem(session=boto3_session._session)  # pylint: disable=protected-access
        full_s3_paths = [f's3://{source_bucket}/{file_path["Key"]}' for file_path in file_paths]
        pq_table = pq.read_table(full_s3_paths, filesystem=f_s, partitioning=None) # type: ignore
        logger.debug(f"Shape of the table {pq_table.shape}")

        try:
            logger.debug(f"Writing to the destination {destination_data_path}")
            pq.write_to_dataset(pq_table,
                                destination_data_path,
                                filesystem=f_s,
                                compression=parquet_file_location.compression,
                                flavor="spark")

            logger.info(f"Done merging, {pq_table.num_rows} rows were written at {destination_data_path}")
            if not parquet_file_location.keep_source_files:
                logger.info('Removing source files')
                logger.debug(f'Removing these files {file_paths}')
                S3Utils.delete_objects(
                    boto_s3_client=s3_client,
                    bucket_name=source_bucket,
                    to_delete=file_paths,
                    logger=logger
                )
        except Exception:
            logger.exception(f'Caught error when trying to merge parquet files: {file_paths}')
            raise

        return pq_table.num_rows

    @staticmethod
    def _remove_files(parquet_file_location, destination_data_path, logger, s3_client):
        """removes files at the destination bucket, assigned in the parquet_file_location"""
        logger.debug(f"Removing all files at {destination_data_path}")
        S3Utils.delete_objects_by_prefix(boto_s3_client=s3_client, bucket_name=parquet_file_location.destination_bucket,
                                         prefix=parquet_file_location.destination_prefix, logger=logger)

    @staticmethod
    def _source_and_destination_not_same(parquet_file_location):
        """checks that source bucket and destination bucket is not the same"""
        if parquet_file_location.source_bucket == parquet_file_location.destination_bucket and \
                parquet_file_location.source_prefix.rstrip('/') == parquet_file_location.destination_prefix.rstrip('/'):
            raise ValueError('Source and destination cannot be the same!')



"""Functions for logging"""
from __future__ import annotations

import itertools
import json
import sys
from typing import Any, Dict

import loguru
from loguru import logger
from loguru._better_exceptions import ExceptionFormatter
from loguru._handler import Message


class LoggingUtils:  # pylint: disable=too-few-public-methods,no-member
    """
    Class that provides useful functions for logging
    """

    exception_formatter = ExceptionFormatter()

    @staticmethod
    def get_logger(name: str, level: str = "INFO") -> loguru.Logger:
        """Returns a logger object"""
        logger.remove()
        logger.configure(extra={"name": name})
        logger.add(LoggingUtils._sink, level=level) # type: ignore

        return logger

    @staticmethod
    def _sink(message: Message) -> None:
        simplified = LoggingUtils._formatter(message.record)
        print(json.dumps(simplified), file=sys.stderr)

    @staticmethod
    def _formatter(record: Dict[str, Any]) -> Dict[str, Any]:
        log_record = {
            "level": record["level"].name,
            "line": record["line"],
            "file": record["file"].name,
            "message": record["message"],
            "time": record["time"].isoformat(),
        }
        log_record.update(record["extra"])

        if record.get("exception"):
            log_record["exception"] = list(
                # chain.from_iterable flattens list of lists
                itertools.chain.from_iterable(
                    e.split("\n")
                    for e in LoggingUtils.exception_formatter.format_exception(
                        *record["exception"]
                    )
                    if e
                )
            )

        return log_record



"""Functionality to fetch/modify info in glue database"""
from __future__ import annotations
from typing import List, Dict

import boto3
import awswrangler as wr
import loguru

from dai_python_commons.dai_error import DaiGlueError


class GlueUtils:  # pylint: disable=too-few-public-methods,no-member
    """Used to fetch/modify info in glue database

    FUNCTIONS
        get_columns_info :func:`~dai_python_commons.GlueUtils.get_columns_info`
        get_partitions :func:`~dai_python_commons.GlueUtils.get_partitions`
        add_partitions :func:`~dai_python_commons.GlueUtils.add_partitions`
    """

    @staticmethod
    def get_columns_info(
        boto3_client, database_name: str, table_name: str
    ) -> List[dict]:
        """
        Gets a list with column information. Each element of the list is a dict including the Name, Type, Comment and
        Parameters
        :param boto3_client: Glue boto3 client
        :param database_name: Name of the database
        :param table_name: Name of the table
        :return: A list of dictionaries including column information. Example:
        [{"Name": "messageid", "Type": "string", "Comment": "The Message id", "Parameters": {}}]
        """
        table_info = boto3_client.get_table(DatabaseName=database_name, Name=table_name)
        columns_info = table_info["Table"]["StorageDescriptor"]["Columns"]

        return columns_info

    @staticmethod
    def get_partitions(
        glue_client,
        database_name: str,
        table_name: str,
        logger: loguru.Logger,
        max_results=100,
    ) -> list:
        """
        Get information about all the partitions of the given table
        :param glue_client: Glue boto3 client with the region set
        :param database_name: Name of the database
        :param table_name: Name of the table
        :param logger: Logger object
        :param max_results: Maximum number of partitions to get when calling the Glue API
        :return: A list of structures including information about the partitions of database_name.table_name. For the
        the exact format see boto3 documentation for glue_client.get_partitions
        """
        logger.info(f"Getting partition information for {database_name}.{table_name}")

        def verify_response(resp):
            if resp["ResponseMetadata"]["HTTPStatusCode"] != 200:
                logger.error(
                    f"Error at getting partitions for {database_name}.{table_name}, response = {resp}"
                )
                raise DaiGlueError("GlueError!")

        response = glue_client.get_partitions(
            DatabaseName=database_name, TableName=table_name, MaxResults=max_results
        )
        verify_response(response)

        all_partitions = response["Partitions"]
        while "NextToken" in response and response["NextToken"]:
            response = glue_client.get_partitions(
                DatabaseName=database_name,
                TableName=table_name,
                MaxResults=max_results,
                NextToken=response["NextToken"],
            )
            verify_response(response)
            all_partitions += response["Partitions"]

        logger.info(f"Found {len(all_partitions)} partitions")

        return all_partitions

    @staticmethod
    def add_partitions(
        session,
        database_name: str,
        table_name: str,
        logger: loguru.Logger,
        partitions2add: List[Dict[str, str]],
    ) -> None:
        """
        Add new partitions to a glue table
        :param session: Boto3 session
        :param database_name: Name of the database
        :param table_name: Name of the table
        :param logger: Logger object
        :param partitions2add: List of partitions to add in the format
        [{
        'Value': ['2022-10-16'],
        'Location': 's3://se-sl-tf-dai-dev-data-refined/tr/ib/0005/assignment/v1/data/atdatetime_dt=2022-10-16'
        }]
        :return: Nothing
        """
        logger.info(
            f"Adding {len(partitions2add)} partitions to {database_name}.{table_name}: {partitions2add}"
        )

        # make sure we have / at the end of the s3 path
        formatted = {}
        for partition in partitions2add:
            loc = (
                partition["Location"]
                if partition["Location"].endswith("/")
                else f"{partition['Location']}/"
            )
            formatted[loc] = partition["Value"]

        wr.catalog.add_parquet_partitions(
            database=database_name,
            table=table_name,
            partitions_values=formatted,
            boto3_session=session,
        )
        logger.info("Partitions added.")

    @staticmethod
    def get_table_location(session: boto3.session.Session, database_name: str, table_name: str) -> str:
        """
        Returns the s3 location of a Glue table. Note that it does not include the partitions paths. Example
        of answer: s3://se-sl-tf-dai-prod-data-refined/tr/ib/0005/apc/v1/data
        :param session: Boto3 session
        :param database_name: Name of the database
        :param table_name: Name of the table
        :return:
        """
        return wr.catalog.get_table_location(
            database=database_name, table=table_name, boto3_session=session
        )



# pylint: skip-file
# File skipped because the refactoring will be complex and requires a new ticket
import datetime
import os
import urllib.parse
from enum import Enum, auto
from io import BytesIO
from typing import Any, Dict, Optional
from urllib.parse import unquote_plus
from zipfile import BadZipFile, ZipFile

from botocore.exceptions import ClientError
from mypy_boto3_s3.service_resource import S3ServiceResource as S3Resource
from mypy_boto3_s3.type_defs import CopyObjectOutputTypeDef, CopySourceTypeDef, ObjectTypeDef

import dai_python_commons.dai_error as dai_error
from dai_python_commons.dai_error import DaiError, S3BucketLocation
from dai_python_commons.event_bridge import EventBridgeS3Event
from dai_python_commons.s3_utils import S3Utils


class ObjectEqualityStrategy(Enum):
    ETAG = auto()
    SIZE = auto()


class S3EventObjectCopy:
    """
    Class that provides functionality to copy objects that triggered an S3 event from one bucket to another
    Useful in lambda functions that have an S3 trigger on PUT operations
    """

    def __init__(self, s3_resource: S3Resource, logger, destination_bucket: str, source_prefix_to_replace: str,
                 destination_prefix_replacement: str, unzip_file: bool, filename_prefix: str = '',
                 remove_processed: bool = False,
                 object_equality_strategy: ObjectEqualityStrategy = ObjectEqualityStrategy.ETAG):
        """
        :param s3_resource: Boto3 resource for S3
        :param logger: The logger
        :param destination_bucket: The name of the destination bucket
        :param source_prefix_to_replace: A part of the prefix in the source bucket that should be replaced with another
        value in the destination.
        :param destination_prefix_replacement: A prefix that should replace source_prefix_to_replace at the destination
        bucket. If for example, we should copy an object source_bucket/A/B/C/x.txt to destination_bucket, and
        source_prefix_to_replace = "/A/B/", destination_prefix_replacement = "E/", then the object will be copied at
        destination_bucket/E/C/x.txt"
        :param unzip_file: Whether the file should be unzipped
        :param filename_prefix: A string that will be put as prefix at the destination, for example:
        "x.txt" --> "<prefix>_x.txt"
        :param remove_processed: Whether the file should be removed from the source bucket after processing. Note that
        this option is not supported when unzip_file = True, or for files larger than 4.9 Gb.
        The file is removed only if the processing was successful
        :param object_equality_strategy: If ETAG or SIZE should be used to measure the equality when removing
        processed objects. ETAG is more accurate but doesn't always work, SIZE is the next best thing.
        """
        self.s3_resource = s3_resource
        self.s3_client = self.s3_resource.meta.client
        self.logger = logger
        self.destination_bucket = destination_bucket
        self.source_prefix_to_replace = source_prefix_to_replace
        self.destination_prefix_replacement = destination_prefix_replacement
        self.unzip_file = unzip_file
        self.filename_prefix = filename_prefix
        self.remove_processed = remove_processed
        self.object_equality_strategy = object_equality_strategy
        if self.remove_processed and self.unzip_file:
            raise dai_error.IncorrectInputValueError(
                {"unzip_file": unzip_file, "remove_unprocessed": remove_processed},
                "The remove_unprocessed option is not supported together with unzip_file.")

    @staticmethod
    def service_exception(response: CopyObjectOutputTypeDef) -> bool:
        """
        Given a response from a boto3 S3 client command, check if a service error occurred

        :param response: Response from an S3 client request
        :return: True if an error occurred, False otherwise
        """
        if "Error" in response:
            return True

        if "ResponseMetadata" in response and "HTTPStatusCode" in response["ResponseMetadata"] and \
                response["ResponseMetadata"]["HTTPStatusCode"] != 200:
            return True

        return False

    @staticmethod
    def extract_field_from_dict(the_dict: dict, field_name: str):
        """
        Extract a specific field from a dictionary

        :param the_dict: dictionary
        :param field_name: field to be extracted
        :return: The value of field_name in the_dict or None if the field is not found in the dictionary
        """
        if the_dict and field_name in the_dict:
            return the_dict[field_name]

        return None

    def are_equal(self, object1_info: ObjectTypeDef, object2_info: ObjectTypeDef) -> bool:
        """
        compares 2 responses from listing s2 objects and asserts on their equality
        """
        if self.object_equality_strategy is ObjectEqualityStrategy.ETAG:
            return S3EventObjectCopy.are_etags_identical(object1_info, object2_info)
        elif self.object_equality_strategy is ObjectEqualityStrategy.SIZE:
            return S3EventObjectCopy.are_objects_same_size(object1_info, object2_info)
        else:
            raise ValueError(f"Unknown object equality strategy: {self.object_equality_strategy}")

    @staticmethod
    def are_etags_identical(object1_info: ObjectTypeDef, object2_info: ObjectTypeDef) -> bool:
        """
        Given two dictionaries as returned by S3 method list_objects_v2, check if their ETags are identical

        :param object1_info: Dictionary as returned by list_objects_v2
        :param object2_info: Dictionary as returned by list_objects_v2
        :return: True if the ETags are identical, False otherwise
        """
        etag1 = object1_info.get("ETag")
        etag2 = object2_info.get("ETag")
        if not etag1 or not etag2 or etag1 != etag2:
            return False

        return True

    @staticmethod
    def are_objects_same_size(object1_info: ObjectTypeDef, object2_info: ObjectTypeDef) -> bool:
        """
        Given two dictionaries as returned by S3 method list_objects_v2, check if they are the same size

        :param object1_info: Dictionary as returned by list_objects_v2
        :param object2_info: Dictionary as returned by list_objects_v2
        :return: True if the objects have the same size, False otherwise
        """
        size1 = object1_info.get("Size")
        size2 = object2_info.get("Size")
        if size1 is None or size2 is None:
            return False

        return size1 == size2

    @staticmethod
    def get_object_lock_retain_until_date(tag_dict: Optional[Dict[str, Any]]) -> Optional[datetime.datetime]:
        """
        Calculates the object lock retain until date based on the *retention_time* tag value

        :param tag_dict: dict that may contain the *retention_time* key
        :return: the "retain until date" or None if no *retention_time* can be found in *destination_tags*
        """

        if tag_dict is not None and "retention_time" in tag_dict:
            nr_days = int(tag_dict["retention_time"])
            if nr_days > 0:
                return datetime.datetime.now() + datetime.timedelta(days=nr_days)
            else:
                raise DaiError("Retention time must be greater than 0 days!")

        return None

    def _unzip_and_copy(self, source_bucket_name: str, source_object: str, destination_prefix_wo_filename: str):
        """
        Unzip a file and copy to destination without persisting to disk

        :param source_bucket_name: The bucket where the zip file is located
        :param source_object: The full name of the zip file in the source bucket
        :param destination_prefix_wo_filename: The prefix in the destination bucket, excluding the filename
        :return:
        """
        self.logger.debug(f"Unzip {source_bucket_name}/{source_object} and copy all files to"
                          f" {self.destination_bucket}/{destination_prefix_wo_filename}")
        # read the zipfile as bytes in memory
        zip_obj = self.s3_resource.Object(bucket_name=source_bucket_name, key=source_object)
        buffer = BytesIO(zip_obj.get()["Body"].read())

        # iterate over the files in the zip and copy them to the destination bucket
        zip_file = ZipFile(buffer)
        for filename in zip_file.namelist():
            # create the full prefix
            new_filename = self._append_filename_prefix(filename)
            destination_prefix = os.path.join(destination_prefix_wo_filename, new_filename)
            # upload the object
            self.logger.debug(f"{filename} -> {destination_prefix}")
            self.s3_client.upload_fileobj(zip_file.open(filename), Bucket=self.destination_bucket,
                                          Key=destination_prefix)

    def _get_object_info(self, bucket: str, prefix: str) -> Optional[ObjectTypeDef]:
        """
        Return metadata information about an object stored in S3

        :param bucket: name of the bucket
        :param prefix: prefix of the object
        :return: A dictionary as returned by list_objects_v2, or None if the object doesn't exist
        """
        result = None
        try:

            response = self.s3_client.list_objects_v2(Bucket=bucket, Prefix=prefix)
            # get first element in Contents list
            result = next(iter(response.get("Contents", [])), None)
        # The boto3 exception mechanisms seems pretty badly documented so I use Exception to be on the safe side
        except Exception as the_exception:
            self.logger.warning(
                f"Unable to get information about object {bucket}/{prefix}, error: {the_exception}")

        return result

    def _remove_processed_object(self, source_bucket: str, source_prefix: str, source_object_info: ObjectTypeDef,
                                 destination_object_info: ObjectTypeDef) -> bool:
        """
        Removes an object that has been processed. It only removes the object if the source and destination etags
        match

        :param source_bucket: The name of the source bucket
        :param source_prefix: The key of the object
        :param source_object_info: Metadata information about the source object as returned by boto3 list_objects_v2
        :param destination_object_info: Metadata information about the destination object as returned by boto3
         list_objects_v2
        :return: True if the object was deleted successfully, False otherwise
        """
        self.logger.debug(f"Attempting to delete processed object {source_bucket}/{source_prefix}")

        source_object = f"{source_bucket}/{source_prefix}"

        # if the objects are not equal the source object is not removed
        if not self.are_equal(source_object_info, destination_object_info):
            self.logger.warning(f"Source and destination objects do not match, the processed object "
                                f"{source_object} will not be removed")
            return False

        # delete the processed file
        try:
            self.s3_client.delete_object(Bucket=source_bucket, Key=source_prefix)
        except Exception as the_exception:
            self.logger.opt(exception=True).warning(f"Unable to remove {source_object}, error: {the_exception}")
            return False
        else:
            self.logger.debug(f"Object {source_object} was deleted")
            return True

    def _get_destination_prefix(self, source_prefix: str, exclude_filename: bool = False) -> str:
        """
        Gets the destination prefix for a given source prefix

        :param source_prefix The source prefix of the object
        :param exclude_filename: Whether the 'filename' should be excluded from the prefix
        :return: The prefix of the object at the destination
        """
        destination_prefix_wo_filename = os.path.dirname(source_prefix)
        if self.source_prefix_to_replace is not None:
            destination_prefix_wo_filename = destination_prefix_wo_filename.replace(
                self.source_prefix_to_replace.strip("/"), self.destination_prefix_replacement.strip("/"))

        if exclude_filename:
            return destination_prefix_wo_filename

        filename = self._append_filename_prefix(os.path.basename(source_prefix))

        return os.path.join(destination_prefix_wo_filename, filename)

    def _append_filename_prefix(self, filename: str):
        """
        If a filename prefix is configured it will be added to the given filename.

        Examples::
            foo.txt     --> <filename_prefix>_foo.txt

        :param filename: the filename that should have the prefix added to it (no 'folder' path)
        :return: the filename with prefix added, if filename prefix configured, else same as given filename
        """
        if self.filename_prefix:
            filename = f'{self.filename_prefix}_{filename}'

        return filename

    def _get_tagging_str(self, tag_dict: Optional[Dict[str, Any]]) -> str:
        """
        :param tag_dict: Dictionary giving tag_name, tag_value
        :return: A string that can be used by boto3 copy commands to tag objects
        """
        return urllib.parse.urlencode(tag_dict) if tag_dict else ""

    def _copy_using_unmanaged_transfer(self,
                                       source_bucket_name: str,
                                       source_object: str,
                                       copy_source_object: CopySourceTypeDef,
                                       object_info: ObjectTypeDef,
                                       destination_prefix: str,
                                       destination_tags: Optional[Dict[str, Any]]):
        """
        Uses client.copy_object to copy an object from one location to another.

        :param source_bucket_name: The name of the source bucket
        :param source_object: Key of the object to be copied
        :param copy_source_object: Dict containing information about the source bucket and key
        :param object_info: Dictionary as returned by list_objects_v2
        :param destination_prefix: The destination prefix
        :param destination_tags: Tags that should be set on the object in the request.
        :return:
        """
        tagging = self._get_tagging_str(destination_tags)

        object_lock_arguments = {}
        if S3Utils.is_raw_bucket(self.destination_bucket) and destination_tags:
            if destination_tags.get("retention_type") in ["shall", "may"]:
                object_lock_arguments = {
                    "ObjectLockMode": "GOVERNANCE",
                    "ObjectLockRetainUntilDate": self.get_object_lock_retain_until_date(destination_tags)
                }
            elif destination_tags.get("retention_type") == "preserve":
                # 52000 weeks are ~1000 years
                object_lock_arguments = {
                    "ObjectLockMode": "GOVERNANCE",
                    "ObjectLockRetainUntilDate": datetime.datetime.now() + datetime.timedelta(weeks=52000)
                }
        response = self.s3_client.copy_object(CopySource=copy_source_object,
                                              Bucket=self.destination_bucket,
                                              Key=destination_prefix,
                                              Tagging=tagging,
                                              TaggingDirective="REPLACE",
                                              **object_lock_arguments
                                              )

        # check if the copying was successful
        s3_bucket_location = S3BucketLocation(source_bucket_name,
                                              source_object,
                                              self.destination_bucket,
                                              destination_prefix)
        if S3EventObjectCopy.service_exception(response):
            raise dai_error.DaiS3CopyObjectError(s3_bucket_location)
        if self.remove_processed:
            destination_object_info = self._get_object_info(self.destination_bucket, destination_prefix)
            if not destination_object_info:
                return
            self._remove_processed_object(source_bucket=source_bucket_name, source_prefix=source_object,
                                          source_object_info=object_info,
                                          destination_object_info=destination_object_info)

    def _copy_using_managed_transfer(self,
                                     copy_source_object: CopySourceTypeDef,
                                     destination_prefix: str,
                                     content_length: Optional[int],
                                     destination_tags: Optional[Dict[str, Any]]):
        """
        Uses client.copy to copy an object from one location to another. This copy is a so called managed transfer.

        client.copy docs:

                "This is a managed transfer which will perform a multipart copy in multiple threads if necessary."

        :param copy_source_object: Dict containing information about the source bucket and key
        :param destination_prefix: The destination prefix
        :param content_length: The size of the object
        :param destination_tags: Tags that should be set on the object in the request.
        :return:
        """
        s3_bucket_location = S3BucketLocation(copy_source_object.get("Bucket", ""),
                                              copy_source_object.get("Key", ""),
                                              self.destination_bucket,
                                              destination_prefix)

        if S3Utils.is_raw_bucket(self.destination_bucket):
            raise dai_error.DaiS3CopyObjectError(
                s3_bucket_location,
                message="Can't copy object due to destination bucket requiring "
                        "object lock which is not supported by this method!"
            )

        tagging = self._get_tagging_str(destination_tags)

        self.s3_client.copy(CopySource=copy_source_object,
                            Bucket=self.destination_bucket,
                            Key=destination_prefix,
                            ExtraArgs={
                                "Tagging": tagging,
                                "TaggingDirective": "REPLACE"
                            })
        # this would require a polling mechanism; there is no plan to have objects so large,
        # so this should be implemented when/if the need arises
        if self.remove_processed:
            if not content_length:
                self.logger.warning("Unable to determine content length, unable to remove "
                                    "{source_bucket_name}/{source_object} ")
            elif content_length is not None and content_length > 4.9e9:
                self.logger.warning("Removal of processed objects for large objects (>= 5Gb) not supported "
                                    "unable to remove {source_bucket_name}/{source_object} ")
            else:
                raise dai_error.DaiS3Error(f"Don't know how to remove processed: {copy_source_object}")

    def _copy_one_object(self, source_bucket_name: str, source_object: str, destination_tags: Optional[Dict[str, Any]] = None):
        """
        Copies one object to the destination bucket

        :param source_bucket_name: Name of the bucket
        :param source_object: Key of the object to be copied
        :param destination_tags: Tags to append to destination file
        :return: Nothing
        :raises dai_error.DaiS3CopyObjectError if the copy_object operation fails
        """

        self.logger.debug(f"Processing {source_bucket_name}/{source_object} ")
        # if needed, unzip the file in memory then copy
        if self.unzip_file:
            destination_prefix_wo_filename = self._get_destination_prefix(source_object, exclude_filename=True)
            self._unzip_and_copy(source_bucket_name, source_object, destination_prefix_wo_filename)
        else:
            # copy the file without unzipping
            copy_source_object = CopySourceTypeDef(Bucket=source_bucket_name, Key=source_object)
            destination_prefix = self._get_destination_prefix(source_object, exclude_filename=False)

            self.logger.debug(f"{source_bucket_name}/{source_object} -> {self.destination_bucket}/{destination_prefix}")

            object_info = self._get_object_info(source_bucket_name, source_object)
            if object_info is not None:
                content_length = object_info.get("Size")
                # use different methods depending on object size
                if not self.remove_processed or content_length is None or content_length > 4.9e9:
                    self._copy_using_managed_transfer(content_length=content_length,
                                                      copy_source_object=copy_source_object,
                                                      destination_prefix=destination_prefix,
                                                      destination_tags=destination_tags)
                else:
                    self._copy_using_unmanaged_transfer(copy_source_object=copy_source_object,
                                                        destination_prefix=destination_prefix,
                                                        object_info=object_info, source_bucket_name=source_bucket_name,
                                                        source_object=source_object, destination_tags=destination_tags)
            else:
                self.logger.warning(
                    f"The object {source_bucket_name}/{source_object} does not exist, perhaps it was already copied?")

        self.logger.debug(f"Finished processing {source_bucket_name}/{source_object} ")

    def copy_objects_s3_event(self, event: Dict[str, Any], destination_tags: Optional[Dict[str, Any]] = None) -> int:
        """
        Copy the object(s) that triggered an S3 event to a destination bucket.

        :param event: The event that triggered the function
        :param destination_tags: The tags that should be added to the object in the destination
        :return: the number of objects copied
        """
        num_records_copied = 0
        for record in event['Records']:
            source_bucket_name = record['s3']['bucket']['name']
            if source_bucket_name.strip() == self.destination_bucket:
                incorrect_inputs = {"destination_bucket_name": self.destination_bucket}
                raise dai_error.IncorrectInputValueError(
                    incorrect_inputs, "The destination bucket cannot be the same as the source bucket.")
            source_object = record['s3']['object']['key']
            try:
                self._copy_one_object(
                    source_bucket_name=source_bucket_name,
                    source_object=unquote_plus(source_object),
                    destination_tags=destination_tags
                )
                num_records_copied += 1
            except (BadZipFile, ClientError) as e:
                self.logger.exception(f"Caught exception: {e}")
                raise

        return num_records_copied

    def copy_objects_event_bridge_s3_event(self, event: EventBridgeS3Event, destination_tags: Optional[Dict[str, Any]] = None):
        source_bucket_name = event.s3_bucket
        # needed to handle special characters
        source_key = unquote_plus(event.s3_key)

        self.logger.info(f"Copying object s3://{source_bucket_name}/{source_key} to bucket {self.destination_bucket}")

        if source_bucket_name == self.destination_bucket:
            incorrect_inputs = {"destination_bucket_name": self.destination_bucket}
            raise dai_error.IncorrectInputValueError(
                incorrect_inputs, "The destination bucket cannot be the same as the source bucket.")

        self._copy_one_object(
            source_bucket_name=source_bucket_name,
            source_object=source_key,
            destination_tags=destination_tags
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dai_python_commons']

package_data = \
{'': ['*']}

install_requires = \
['boto3-stubs[glue,s3]==1.26.99', 'boto3==1.26.99', 'loguru>=0.6.0,<0.7.0']

extras_require = \
{'data-consolidation': ['pyarrow==3.0.0', 's3fs==0.4.2'],
 'glue': ['awswrangler==2.18.0', 'cryptography==38.0.1']}

setup_kwargs = {
    'name': 'dai-python-commons',
    'version': '4.12.0',
    'description': 'Collection of small python utilities useful for lambda functions or glue jobs. By the Stockholm Public Transport Administration.',
    'long_description': 'None',
    'author': 'None',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)



"""
This module defines the exceptions used throughout this package
"""
from dataclasses import dataclass
from typing import Optional


@dataclass
class S3BucketLocation:
    """
    Class that contains information about objects location in s3.

    args:
        source_bucket (str): The bucket where objects are located.
        source_prefix (str): The prefix (ie 'folder') where the objects are located.
        destination_bucket (str): The bucket where the copy-to destination is located
        destination_prefix (str): The prefix (ie 'folder') where the copy-to destination is located
    """

    source_bucket: str
    source_prefix: str
    destination_bucket: str
    destination_prefix: str


class DaiError(Exception):
    """Base class for exceptions in DAI."""

class DaiGlueError(DaiError):
    """Base class for exceptions related to Glue."""

class DaiS3Error(DaiError):
    """Base class for exceptions related to S3."""


class DaiS3CopyObjectError(DaiS3Error):
    """Class for exceptions related to copying objects in S3."""
    def __init__(self, s3_bucket_location: S3BucketLocation, message=None):  # pylint: disable=super-init-not-called
        self.source = f"s3://{s3_bucket_location.source_bucket}/{s3_bucket_location.source_prefix}"
        self.destination = f"s3://{s3_bucket_location.destination_bucket}/{s3_bucket_location.destination_prefix}"
        self.message = message

    def __repr__(self):
        ret_str = f"source: {self.source}, destination: {self.destination}"
        if self.message:
            ret_str += f"\nError message: {self.message}"

        return ret_str

    def __str__(self):
        ret_str = f"Error at copying: {self.source} -> {self.destination}"
        if self.message:
            ret_str += f"\nError message: {self.message}"

        return ret_str


class DaiS3DeleteObjectError(DaiS3Error):
    """Class for exceptions related to deleting objects in S3."""

    def __init__(self, s3_bucket: str, error_infos: list, message: Optional[str] = None):  # pylint: disable=super-init-not-called
        self.bucket = s3_bucket
        self.error_infos = error_infos
        self.message = message

    def __repr__(self):
        ret_str = f"Bucket: {self.bucket}, Error info: {self.error_infos}"
        if self.message:
            ret_str += f"\nError message: {self.message}"

        return ret_str

    def __str__(self):
        ret_str = f"Error at deleting objects in bucket {self.bucket}."
        if self.message:
            ret_str += f"\nError message: {self.message}"
        ret_str += "\nAdditional information about failed removals: \n"
        as_str = [f"{error_info}" for error_info in self.error_infos]
        ret_str += '\n'.join(as_str)

        return ret_str


class DaiInputError(DaiError):
    """Base class for exceptions related to input data in DAI."""


class MissingInputError(DaiInputError):
    """Exception raised when missing critical input information.

    Attributes:
        missing_inputs -- list of missing inputs
        message -- explanation of the error
    """

    def __init__(self, inputs_list, message):  # pylint: disable=super-init-not-called
        self.inputs_list = inputs_list
        self.message = message

    def __repr__(self):
        ret_str = f"Missing inputs: {self.inputs_list}"
        ret_str += f"\nError message: {self.message}"

        return ret_str

    def __str__(self):
        ret_str = f"The following inputs are missing: {self.inputs_list}"
        ret_str += f"\nError message: {self.message}"

        return ret_str


class IncorrectInputValueError(DaiInputError):
    """Exception raised when incorrect input(s) provided.

    Attributes:
        input_dict -- dictionary providing the name and values of the inputs containing incorrect values.
        message -- explanation of the error
    """

    def __init__(self, incorrect_inputs_dict, message):  # pylint: disable=super-init-not-called
        self.incorrect_inputs_dict = incorrect_inputs_dict
        self.message = message

    def __repr__(self):
        ret_str = f"Incorrect inputs: {self.incorrect_inputs_dict}"
        ret_str += f"\nError message: {self.message}"

        return ret_str

    def __str__(self):
        ret_str = f"The following input(s) received incorrect value(s): {self.incorrect_inputs_dict}"
        ret_str += f"\nError message: {self.message}"

        return ret_str
